(*
Copyright 2010 Christian Sternagel, Ren Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory CetaVerifier
imports CPF_Parser
begin

abbreviation drop_white_space :: "string \<Rightarrow> string" 
where "drop_white_space \<equiv> filter (\<lambda> c. \<not> c mem wspace)"

definition eqWhiteSpace :: "string \<Rightarrow> string \<Rightarrow> bool" ("_ =w _" 50)
where "s1 =w s2 \<equiv> debug ''0''
  (''comparing whether parsed input corresponds to real input'')
  (drop_white_space s1 = drop_white_space s2)"

fun search_len :: "string \<Rightarrow> string \<Rightarrow> nat \<Rightarrow> (string \<times> string)option"
where "search_len t s n = (if take n t = s then Some ([],drop n t) 
                else (case t of Nil \<Rightarrow> None | Cons h tt \<Rightarrow> (case search_len tt s n of None \<Rightarrow> None | Some (bef,aft) \<Rightarrow> Some (h # bef,aft))))"

lemma search_len_sound:
  fixes t s before after::string
  assumes "search_len t s n = Some (before,after)"
  shows "t = (before @ s @ after)"
using assms
proof (induct t arbitrary: before)
  case Nil
  thus ?case 
    by (cases s, auto)
next
  case (Cons h tt)
  show ?case 
  proof (cases "take n (h # tt) = s")
    case True
    with Cons show ?thesis by auto 
  next
    case False with `search_len (h # tt) s n = Some (before,after)` 
    have res: "(case search_len tt s n of None \<Rightarrow> None | Some (bef,aft) \<Rightarrow> Some (h # bef, aft)) = Some (before,after)" by auto
    show ?thesis 
    proof (cases "search_len tt s n")
      case None with res show ?thesis by auto
    next
      case (Some befaft)
      show ?thesis proof (cases befaft)
	case (Pair bef aft)
	with Some res have "search_len tt s n = Some (bef,aft) \<and> before = h # bef \<and> after = aft" by auto
	with Cons show ?thesis by auto
      qed
    qed
  qed
qed

declare search_len.simps[simp del]    

abbreviation search :: "string \<Rightarrow> string \<Rightarrow> (string \<times> string)option"
where "search t s \<equiv> search_len t s (length s)"

fun starts_with :: "string \<Rightarrow> string \<Rightarrow> bool"
where "starts_with t s = (if (t \<noteq> '''' \<and> hd t mem wspace) then starts_with (tl t) s else 
           case search t s of None \<Rightarrow> False | Some _ \<Rightarrow> True)"

abbreviation s_proof :: string where "s_proof \<equiv> ''<proof>''"
abbreviation s_input :: string where "s_input \<equiv> ''<input>''"
abbreviation e_input :: string where "e_input \<equiv> ''</input>''"


fun extract_input_proof :: "string \<Rightarrow> (string \<times> string \<times> string) option"
where "extract_input_proof s = (
   case search s s_input of
      None \<Rightarrow> None
   | Some (_,minput) \<Rightarrow> (
   case search minput e_input of 
      None \<Rightarrow> None
   | Some (the_input,s') \<Rightarrow> (
   case search s' s_proof of
      None \<Rightarrow> None
   | Some (version,prf) \<Rightarrow> Some (the_input,version,prf))))" 


lemma extract_input_proof_sound: 
  fixes append :: "string \<Rightarrow> string \<Rightarrow> string" (infixr "<@>" 65)
  defines [simp]: "append \<equiv> List.append"
  assumes res: "extract_input_proof s = Some (the_input,version,the_proof)"
  shows "\<exists> bef. s = (bef <@> ''<input>'' <@> the_input <@> ''</input>'' <@> version <@> ''<proof>'' <@> the_proof)"
proof (cases "search s s_input")
  case None
  with res show ?thesis by auto
next
  case (Some bef_minput)
  show ?thesis
  proof (cases bef_minput)
    case (Pair bef minput)
    with Some have s1: "search s s_input = Some (bef, minput)" by auto
    show ?thesis 
    proof (cases "search minput e_input")
      case None
      with res s1 show ?thesis by auto
    next
      case (Some input_sp)
      show ?thesis
      proof (cases input_sp)
	case (Pair tthe_input s')
	with Some s1 have s2: "search minput e_input = Some (tthe_input, s')" by auto
	show ?thesis
	proof (cases "search s' s_proof")
	  case None 
	  with res s1 s2 show ?thesis by auto
	next
	  case (Some vers_proof)
	  show ?thesis 
	  proof (cases vers_proof)
	    case (Pair _ _)
	    with Some s1 s2 res have s3: "search s' s_proof = Some (version, the_proof)" by auto		
            from res s1 s2 s3 have input: "tthe_input = the_input" by simp
	    from s1 have sp1: "s = (bef @ s_input @ minput)" by (rule search_len_sound)
	    from s2 have sp2: "minput = (tthe_input @ e_input @ s')" by (rule search_len_sound)
	    from s3 have sp3: "s' = (version @ s_proof @ the_proof)" by (rule search_len_sound)            
            show ?thesis
              by (rule exI[of _ bef], simp add: sp1 sp2 sp3 input)
          qed
        qed
      qed
    qed
  qed
qed

  
fun certifyProof :: "string \<Rightarrow> certificationResult"
where "certifyProof problemString  = (
  case parseCertProblem problemString of
     Inl err \<Rightarrow> Error (''error while parsing'' @ [newline] @ err)
   | Inr cp \<Rightarrow> case extract_input_proof problemString of
         None \<Rightarrow> Error ''could not find input and proof in string for comparison with parsed result''
       | Some (the_input, version, the_proof) \<Rightarrow> (
             if (the_input =w (to_string (xml cp))) 
                 then case (certifyCertProblem cp) of
                         Certified tag \<Rightarrow> if starts_with the_proof tag 
                                                then Certified tag 
                                                else Error (''proven property '' @ tag @ '' does not correspond to proof in input string: '' @ (take 20 the_proof)) 
                      |  Error e \<Rightarrow> Error e
                 else Error (concat [''parsed problem does not correspond to input'', [newline], ''input:'',[newline], drop_white_space the_input,[newline],[newline],''parsed:'',[newline],drop_white_space (to_string (xml cp)), [newline]])
  ))" 


text {*
if the certifier gives an answer then
- the problem string contains an input and a proof of some desired property
- moreover, there is a certification problem with some input and some proof where the pretty printer delivers exactly the same input as in the problem string
  and the proof-goals are identical
- and for the input the desired property for which the proof has been given is established
*}
lemma certifyProofSound: 
  fixes append :: "string \<Rightarrow> string \<Rightarrow> string" (infixr "<@>" 65)
  defines [simp]: "append \<equiv> List.append"
  assumes answer: "certifyProof problemString = Certified answer"
  shows "\<exists> (cp :: (string, nat list, string)certificationProblem) the_problem the_input version the_proof. 
               problemString = the_problem <@> ''<input>'' <@> the_input <@> ''</input>'' <@> version <@> ''<proof>'' <@> the_proof \<and>
                the_input =w to_string (xml (cp)) \<and> starts_with the_proof (xml_tag cp) \<and> xml_tag cp = answer \<and> desired_property cp"
proof (cases "parseCertProblem problemString")
  case (Inl err)
  with answer show ?thesis by auto
next
  case (Inr cp)
  show ?thesis 
  proof (cases "extract_input_proof problemString")
    case None
    with Inr answer show ?thesis by auto
  next
    case (Some triple)
    show ?thesis
    proof (cases triple)
      case (fields the_input version the_proof)
      with Some have oSome: "extract_input_proof problemString = Some (the_input, version, the_proof)" by simp
      from extract_input_proof_sound[OF oSome] obtain bef where problem: "problemString = bef @ s_input @ the_input @ e_input @ version @ s_proof @ the_proof" ..
      show ?thesis
      proof (cases "the_input =w to_string (xml cp)")
	case False
	with answer Inr oSome show ?thesis by auto
      next
	case True note oTrue = this
        show ?thesis
        proof (cases "certifyCertProblem cp")
          case Error
          with oSome Inr True answer show ?thesis by simp
        next
          case (Certified answ)
          show ?thesis 
          proof (cases "starts_with the_proof answ")
            case False
            with oSome Inr True answer Certified show ?thesis by simp
          next
            case True
	    have "certifyProof problemString = Certified answ" by (simp add: Inr oSome[simplified] oTrue[simplified], simp only: Certified certificationResult.cases True, simp)
            with answer have answer: "answ = answer" by simp
            show ?thesis
            proof (intro exI conjI)
              show "problemString = bef <@> s_input <@> the_input <@> e_input <@> version <@> s_proof <@> the_proof"
                by (simp add: problem)
            next
              show "the_input =w to_string (xml cp)"
                by (rule oTrue)
            next
              show "starts_with the_proof (xml_tag cp)"
	        using certifyCertProblemSound[OF Certified] answer True by simp
            next
              show "xml_tag cp = answer"
	        using certifyCertProblemSound[OF Certified] answer True by simp
            next
              show "desired_property cp"
                using certifyCertProblemSound[OF Certified] ..
            qed
          qed
        qed
      qed
    qed
  qed
qed


end
