(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory SubtermCriterion imports DpFramework begin

subsection {* Projections *}

text {*
A projection @{term "proj_term p t"} maps a term with root symbol @{term "f"}
to its @{term "p f"}-th argument or does not change the term at all (if
@{term "p f"} is not an argument postion of @{term "f"}. Variables are not changed.
*}
types
  'f proj = "'f \<Rightarrow> nat"

fun
  proj_term :: "'f proj \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term"
where
  "proj_term p (Var x) = Var x" |
  "proj_term p (Fun f ts) = (if p f < length ts
     then ts ! (p f)
     else Fun f ts)"

lemma proj_term_subst_distr[simp]:
  assumes "\<pi> f < length ss" shows "proj_term \<pi> (Fun f ss) \<cdot> \<sigma> = proj_term \<pi> (Fun f ss \<cdot> \<sigma>)"
using assms by auto

lemma subterm_preserves_SN:
  assumes SN: "SN_elt (rstep R) t" and supt: "(t,s) \<in> supt"
  shows "SN_elt (rstep R) s"
proof -
  from supt have "t \<unrhd> s" unfolding supt_def suptp_iff_supteqp_neq by simp
  thus ?thesis using SN_imp_SN_subt[OF SN, of s] by simp
qed

lemma SN_elt_rstep_imp_SN_elt_supt_union_rstep:
  assumes "SN_elt (rstep R) t"
  shows "SN_elt (supt \<union> rstep R) t" (is "SN_elt ?S t")
proof (rule ccontr)
  assume "\<not> SN_elt ?S t"
  then obtain s where ini: "s 0 = t" and seq: "\<forall>i. (s i, s (Suc i)) \<in> ?S"
    unfolding SN_elt_def by auto
  have SN: "\<forall>i. SN_elt (rstep R) (s i)"
  proof
    fix i show "SN_elt (rstep R) (s i)"
    proof (induct i)
      case 0 show ?case using assms unfolding `s 0 = t` .
    next
      case (Suc i)
      from seq have "(s i, s (Suc i)) \<in> ?S" by simp
      thus ?case
      proof
        assume "(s i, s (Suc i)) \<in> supt"
        from subterm_preserves_SN[OF Suc this] show ?thesis .
      next
        assume "(s i, s (Suc i)) \<in> rstep R"
        from step_preserves_SN_elt[OF this Suc] show ?thesis .
      qed
    qed
  qed
  have "\<not> (\<exists>S. \<forall>i. (S i, S (Suc i)) \<in> SN_rel_supt (rstep R))"
    using SN_SN_rel_supt_rstep unfolding SN_defs by auto
  moreover have "\<forall>i. (s i, s (Suc i)) \<in> SN_rel_supt (rstep R)"
  proof
    fix i
    from SN have SN: "SN_elt (rstep R) (s i)" by simp
    from seq have "(s i, s (Suc i)) \<in> supt \<union> rstep R" by simp
    thus "(s i, s (Suc i)) \<in> SN_rel_supt (rstep R)"
      unfolding SN_rel_supt_def SN_rel_def using SN by auto
  qed
  ultimately show False by auto
qed

lemma supteq_proj_term: "t \<unrhd> proj_term p t" by (cases t) auto

lemma proj_term_subst_apply_term_distrib[simp]:
  assumes "\<not> is_var t"
  shows "proj_term p t \<cdot> \<sigma> = proj_term p (t \<cdot> \<sigma>)"
using assms by (cases t) simp_all

lemma rstep_proj_term:
  fixes p :: "'f proj"
  defines [simp]: "\<pi> t \<equiv> proj_term p t"
  assumes "\<not> defined R (the (root s))"
    and "wf_trs R"
    and "(s, t) \<in> rstep R"
  shows "(\<pi> s, \<pi> t) \<in> rstep R \<union> Id"
proof -
  from rstep_imp_Fun[OF `wf_trs R` `(s, t) \<in> rstep R`]
    obtain f ss where s: "s = Fun f ss" by best
  from `(s, t) \<in> rstep R` have "(Fun f ss, t) \<in> rstep R" by (simp add: s)
  with `wf_trs R` show ?thesis
  proof (cases rule: rstep_cases_Fun')
    case (root ls r \<sigma>)
    with `\<not> defined R (the (root s))` show ?thesis by (auto simp: s defined_def)
  next
    case (nonroot i u)
    let ?ss1 = "take i ss"
    let ?ss2 = "drop (Suc i) ss"
    let ?C = "More f ?ss1 \<box> ?ss2"
    let ?ss = "?ss1 @ u # ?ss2"
    show ?thesis proof (cases "p f = i")
      case True
      have proj_s: "\<pi> s = ss ! i" using True nonroot by (simp add: s)
      from nth_append_take[of i ss u] and `i < length ss`
        have proj_t: "\<pi> t = u" unfolding nonroot using True by simp
      from `(ss ! i, u) \<in> rstep R` show ?thesis unfolding proj_s proj_t ..
    next
      case False show ?thesis
      proof (cases "p f < i")
        case True
        with nonroot have "p f < length ss" by simp
        hence proj_s: "\<pi> s = ss ! p f" by (simp add: s)
        from True and `p f < length ss` have "p f < length ?ss1" by simp
        hence "\<pi> t = ?ss1 ! p f" unfolding nonroot using nth_append[of "?ss1" _ "p f"] by auto
        with nth_take[OF True, of ss]
          have proj_t: "\<pi> t = ss ! p f" by simp
        show ?thesis unfolding proj_s proj_t by simp
      next
        case False
        with `p f \<noteq> i` have "p f > i" by simp
        show ?thesis proof (cases "p f < length ss")
          case False
          hence proj_s: "\<pi> s = s" unfolding s by simp
          from nonroot have "length ?ss = length ss" by simp
          with False have "\<not> p f < length ?ss" by simp
          hence proj_t: "\<pi> t = t" unfolding nonroot by simp
          from `(s, t) \<in> rstep R` show ?thesis unfolding proj_s proj_t ..
        next
          case True
          hence "p f \<le> length ss" by simp
          from `i < length ss` have "i \<le> length ss" by simp
          from nth_append_take_drop_is_nth_conv[OF `p f \<le> length ss` `i \<le> length ss` `p f \<noteq> i`]
            have "?ss ! p f = ss ! p f" .
          hence proj_s: "\<pi> s = ss ! p f"
            and proj_t: "\<pi> t = ss ! p f"
            using `p f < length ss` by (simp_all add: nonroot s)
          show ?thesis unfolding proj_s proj_t by simp
        qed
      qed
    qed
  qed
qed

lemma rstep_preserves_undefined_root:
  assumes "\<not> defined R (the (root s))" and "wf_trs R" and "(s, t) \<in> rstep R"
  shows "\<not> defined R (the (root t))"
proof -
  from rstep_imp_Fun[OF `wf_trs R` `(s, t) \<in> rstep R`]
    obtain f ss where s: "s = Fun f ss" by best
  from `(s, t) \<in> rstep R` have "(Fun f ss, t) \<in> rstep R" by (simp add: s)
  with `wf_trs R` show ?thesis proof (cases rule: rstep_cases_Fun')
    case (root ls r \<sigma>)
    with `\<not> defined R (the (root s))` show ?thesis by (auto simp: s defined_def)
  next
    case (nonroot i u)
    from `\<not> defined R (the (root s))` show ?thesis unfolding s nonroot by simp
  qed
qed

lemma rsteps_preserve_undefined_root:
  assumes "(s, t) \<in> (rstep R)^*" and "\<not> defined R (the (root s))" and "wf_trs R"
  shows "\<not> defined R (the (root t))"
using assms by (induct) (auto simp: rstep_preserves_undefined_root[OF _ `wf_trs R`])

lemma rsteps_proj_term:
  assumes "(s, t) \<in> (rstep R)^*" and "\<not> defined R (the (root s))" and "wf_trs R"
  shows "(proj_term p s, proj_term p t) \<in> (rstep R)^*"
using assms proof (induct)
  case base show ?case by simp
next
  case (step u v)
  hence seq: "(proj_term p s, proj_term p u) \<in> (rstep R)^*" by simp
  from rsteps_preserve_undefined_root[OF `(s, u) \<in> (rstep R)^*` `\<not> defined R (the (root s))` `wf_trs R`]
    have "\<not> defined R (the (root u))" .
  from rstep_proj_term[OF `\<not> defined R (the (root u))` `wf_trs R`] and `(u, v) \<in> rstep R`
    have "(proj_term p u, proj_term p v) \<in> rstep R \<union> Id" by simp
  with `(proj_term p s, proj_term p u) \<in> (rstep R)^*`
    show "(proj_term p s, proj_term p v) \<in> (rstep R)^*" by auto
qed

lemma supt_rstep_stable:
  assumes "(s, t) \<in> op \<rhd> \<union> rstep R"
  shows "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> op \<rhd> \<union> rstep R"
using assms proof
  assume "(s, t) \<in> op \<rhd>" show ?thesis
  proof (rule UnI1)
    from `(s, t) \<in> op \<rhd>` show "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> op \<rhd>" by (rule suptp_stable[to_set])
  qed
next
  assume "(s, t) \<in> rstep R" show ?thesis
  proof (rule UnI2)
    from `(s, t) \<in> rstep R` show "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> rstep R" ..
  qed
qed

lemma supt_rstep_trancl_stable:
  assumes "(s, t) \<in> (op \<rhd> \<union> rstep R)^+"
  shows "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (op \<rhd> \<union> rstep R)^+"
using assms proof (induct)
  case (base u)
  hence "(s \<cdot> \<sigma>, u \<cdot> \<sigma>) \<in> op \<rhd> \<union> rstep R" by (rule supt_rstep_stable)
  thus ?case ..
next
  case (step u v)
  from `(s \<cdot> \<sigma>, u \<cdot> \<sigma>) \<in> (op \<rhd> \<union> rstep R)^+`
    and supt_rstep_stable[OF `(u, v) \<in> op \<rhd> \<union> rstep R`, of "\<sigma>"]
    show ?case ..
qed

lemma supt_rstep_rtrancl_stable:
  assumes "(s, t) \<in> (op \<rhd> \<union> rstep R)^*"
  shows "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (op \<rhd> \<union> rstep R)^*"
using assms proof (induct)
  case base thus ?case ..
next
  case (step u v)
  from `(s, u) \<in> (op \<rhd> \<union> rstep R)^*` and `(u, v) \<in> op \<rhd> \<union> rstep R`
    have "(s, v) \<in> (op \<rhd> \<union> rstep R)^+" by (rule rtrancl_into_trancl1)
  from trancl_into_rtrancl[OF supt_rstep_trancl_stable[OF this]]
    show ?case .
qed

lemma not_is_var_Fun_conv: "(\<not> is_var t) = (\<exists>f ts. t = Fun f ts)" by auto

lemma wf_trs_def':
  "wf_trs R = (\<forall>(l, r)\<in>R. \<not> is_var l \<and> vars_term r \<subseteq> vars_term l)"
by (rule iffI) (auto simp: wf_trs_def not_is_var_Fun_conv)

definition
  wf_rule :: "('f, 'v) rule \<Rightarrow> bool"
where
  "wf_rule r \<equiv> \<not> is_var (fst r) \<and> vars_term (snd r) \<subseteq> vars_term (fst r)"

definition
  wf_rules :: "('f, 'v) trs \<Rightarrow> ('f, 'v) trs"
where
  "wf_rules R \<equiv> {r. r \<in> R \<and> wf_rule r}"

lemma wf_trs_wf_rules[simp]: "wf_trs (wf_rules R)"
unfolding wf_trs_def' wf_rules_def wf_rule_def split_def by simp

lemma wf_rules_subset[simp]: "wf_rules R \<subseteq> R"
unfolding wf_rules_def by auto

lemma SN_elt_imp_rstep_wf_rules:
  assumes "SN_elt (rstep R) s" and "(s, t) \<in> rstep R"
  shows "(s, t) \<in> rstep (wf_rules R)"
using `(s, t) \<in> rstep R` proof
  fix l r C \<sigma> assume "(l, r) \<in> R"
    and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
  show ?thesis proof (cases "(l, r) \<in> wf_rules R")
    case True thus ?thesis unfolding s t by auto
  next
    case False
    with `(l, r) \<in> R` have "is_var l \<or> (\<exists>x. x \<in> vars_term r - vars_term l)"
      unfolding wf_rules_def wf_rule_def by auto
    thus ?thesis
    proof
      assume "is_var l"
      then obtain x where l: "l = Var x" by auto
      let ?\<sigma> = "{x \<leftarrow> r}"
      let ?S = "\<lambda>i. C\<langle>Var x \<cdot> ?\<sigma>^i \<cdot> \<sigma>\<rangle>"
      have "?S 0 = s" unfolding s l by simp
      moreover have "\<forall>i. (?S i, ?S (Suc i)) \<in> rstep R"
      proof
        fix i
        from rstep.ctxt[OF rstep.subst[OF rstep.id[OF `(l, r) \<in> R`], of "?\<sigma>^i \<circ>s \<sigma>"], of C]
          show "(?S i, ?S (Suc i)) \<in> rstep R" by (simp add: l subst_def)
      qed
      ultimately have "\<not> SN_elt (rstep R) s" unfolding SN_defs by best
      with assms show ?thesis by simp
    next
      assume "\<exists>x. x \<in> vars_term r - vars_term l"
      then obtain x where "x \<in> vars_term r - vars_term l"
        and empty: "vars_term l \<inter> {x} = {}" by auto
      hence "r \<unrhd> Var x" by (induct r) auto
      then obtain D where r: "r = D\<langle>Var x\<rangle>" by (rule supteqp_ctxt_E)
      let ?\<sigma> = "{x \<leftarrow> l}"
      let ?S = "\<lambda>i. C\<langle>((D \<cdot>c ?\<sigma>)^i)\<langle>l\<rangle> \<cdot> \<sigma>\<rangle>"
      from subst_apply_id'[OF empty, of "?\<sigma>"]
        have l: "l \<cdot> ?\<sigma> = l" by simp
      have "?S 0 = s" unfolding s by simp
      moreover have "\<forall>i. (?S i, ?S (Suc i)) \<in> rstep R"
      proof
        fix i
        from rstepI[OF `(l, r) \<in> R`, of _ "(D \<cdot>c ?\<sigma>)^i" "?\<sigma>"]
          have "(((D \<cdot>c ?\<sigma>)^i)\<langle>l\<rangle>, ((D \<cdot>c ?\<sigma>)^i \<circ>c (D \<cdot>c ?\<sigma>)^Suc 0)\<langle>l \<cdot> ?\<sigma>\<rangle>) \<in> rstep R"
          unfolding r l subst_apply_term_ctxt_apply_distrib by (simp add: subst_def)
        from rstep.ctxt[OF rstep.subst[OF this], of C \<sigma>]
          show "(?S i, ?S (Suc i)) \<in> rstep R"
          unfolding ctxt_power_compose_distr[symmetric] by (simp add: l)
      qed
      ultimately have "\<not> SN_elt (rstep R) s" unfolding SN_defs by best
      with assms show ?thesis by simp
    qed
  qed
qed

lemma SN_elt_imp_rsteps_wf_rules:
  assumes "(s, t) \<in> (rstep R)^*" and "SN_elt (rstep R) s"
  shows "(s, t) \<in> (rstep (wf_rules R))^*"
using `(s, t) \<in> (rstep R)^*` proof (induct)
  case base show ?case ..
next
  case (step u v)
  from steps_preserve_SN_elt[OF `(s, u) \<in> (rstep R)^*` `SN_elt (rstep R) s`]
    have "SN_elt (rstep R) u" .
  from SN_elt_imp_rstep_wf_rules[OF this `(u, v) \<in> rstep R`]
    have "(u, v) \<in> rstep (wf_rules R)" .
  with `(s, u) \<in> (rstep (wf_rules R))^*` show ?case ..
qed

lemma SN_elt_rstep_proj_term:
  assumes "SN_elt (rstep R) s" and "\<not> defined R (the (root s))" and "(s, t) \<in> rstep R"
  shows "(proj_term p s, proj_term p t) \<in> rstep R \<union> Id"
proof -
  from assms have undef: "\<not> defined (wf_rules R) (the (root s))"
    using wf_rules_subset[of R] unfolding defined_def wf_rules_def by auto
  from SN_elt_imp_rstep_wf_rules[OF `SN_elt (rstep R) s` `(s, t) \<in> rstep R`]
    have "(s, t) \<in> rstep (wf_rules R)" .
  from rstep_proj_term[OF undef wf_trs_wf_rules this]
    show ?thesis using wf_rules_subset by blast
qed

lemmas rstep_wf_rules_subset = rstep_mono[OF wf_rules_subset]

lemma SN_elt_rsteps_proj_term:
  assumes "(s, t) \<in> (rstep R)^*" and "SN_elt (rstep R) s" and "\<not> defined R (the (root s))"
  shows "(proj_term p s, proj_term p t) \<in> (rstep R)^*"
proof -
  from assms have undef: "\<not> defined (wf_rules R) (the (root s))"
    using wf_rules_subset[of R] unfolding defined_def wf_rules_def by auto
  from SN_elt_imp_rsteps_wf_rules[OF `(s, t) \<in> (rstep R)^*` `SN_elt (rstep R) s`]
    have "(s, t) \<in> (rstep (wf_rules R))^*" .
  from rsteps_proj_term[OF this undef wf_trs_wf_rules, of p]
    show ?thesis
  proof (induct)
    case base thus ?case ..
  next
    case (step u v)
    with rstep_wf_rules_subset[of R] have "(u, v) \<in> rstep R" by blast
    with `(proj_term p s, u) \<in> (rstep R)^*` show ?case ..
  qed
qed

text {*
Lhss and rhss of @{text "P"} are nonvariable terms with roots that are
not defined in @{text "R"}. The projections of the removed rules need
to be in the relation @{term "(op \<rhd> \<union> rstep R)^+"}. The projections of
all other rules need to be equal. This extended version of the subterm
criterion (where now also identity projections and rewrite steps are
possible) is due to the A3PAT team (TODO: add reference). However,
extending our existing proof, was straight-forward, since internally we
already used exactly this relation, but just did not realize that we could
lift it to the main result.
*}
definition
  subterm_criterion_cond :: "'f proj \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow> bool"
where
  "subterm_criterion_cond p P P' R \<equiv>
     (\<forall>(l, r)\<in>P. \<not> is_var l \<and> \<not> is_var r \<and> \<not> defined R (the (root r))) \<and>
     (\<forall>(l, r)\<in>P - P'. (proj_term p l, proj_term p r) \<in> (op \<rhd> \<union> rstep R)^+) \<and>
     (\<forall>(l, r)\<in>P'. proj_term p l = proj_term p r)"

definition
  subterm_criterion_proc ::
    "'f proj \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow> bool"
where
  "subterm_criterion_proc p P R P' R' \<equiv> subterm_criterion_cond p P P' R \<and> R = R'"

fun
  shift_by :: "(nat \<Rightarrow> nat) \<Rightarrow> nat \<Rightarrow> nat"
where
  "shift_by f 0 = 0" |
  "shift_by f (Suc i) = Suc (f (shift_by f i))"

theorem ci_subterm_criterion_proc:
  "ci_proc (subterm_criterion_proc p)"
unfolding ci_proc_def
proof (intro impI allI)
  fix P R P' R' assume proc: "subterm_criterion_proc p P R P' R'"
  hence "R = R'" and "R' \<subseteq> R" unfolding subterm_criterion_proc_def by auto
  show "R' \<subseteq>ci R \<and> (\<forall>s t \<sigma>.
    min_ichain (P, R) s t \<sigma> \<longrightarrow> (\<exists>i. ichain (P', R') (shift s i) (shift t i) (shift \<sigma> i)))"
  proof (intro conjI allI impI)
    from ci_subsetI[OF `R' \<subseteq> R`] show "R' \<subseteq>ci R" .
  next
    fix s t \<sigma> assume mchain: "min_ichain (P, R) s t \<sigma>"
    let ?\<pi> = "\<lambda>t. proj_term p t"
    let ?Q = "P - P'"
    let ?P' = "P' \<inter> P"
    have "P = ?Q \<union> ?P'" by blast
     from proc[unfolded subterm_criterion_proc_def, THEN conjunct1]
      have cond: "\<forall>(l, r)\<in>P. \<not> is_var l \<and> \<not> is_var r \<and> \<not> defined R (the (root r))"
      and strict: "\<forall>(l, r)\<in>?Q. (?\<pi> l, ?\<pi> r) \<in> (op \<rhd> \<union> rstep R)^+"
      and weak: "\<forall>(l, r)\<in>?P'. ?\<pi> l = ?\<pi> r"
      unfolding subterm_criterion_cond_def by blast+
    show "\<exists>i. ichain (P', R') (shift s i) (shift t i) (shift \<sigma> i)"
    proof -
      from mchain
        have rsteps: "\<forall>i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep R)^*" by simp
      have "(\<exists>j. \<forall>i\<ge>j. (s i, t i) \<in> ?P') \<or> (\<forall>i. \<exists>j\<ge>i. (s j, t j) \<notin> ?P')" by blast
      thus ?thesis proof
        assume "\<exists>i. \<forall>j\<ge>i. (s j, t j) \<in> ?P'"
        then obtain i where tail: "\<forall>j\<ge>i. (s j, t j) \<in> ?P'" by auto
        let ?s = "shift s i"
        let ?t = "shift t i"
        let ?\<sigma> = "shift \<sigma> i"
        from tail have "\<forall>j. (?s j, ?t j) \<in> ?P'" by auto
        moreover from rsteps
          have "\<forall>j. (?t j \<cdot> ?\<sigma> j, ?s (Suc j) \<cdot> ?\<sigma> (Suc j)) \<in> (rstep R)^*" by auto
        ultimately show ?thesis by (auto simp: `R = R'`)
      next
        assume "\<forall>i. \<exists>j\<ge>i. (s j, t j) \<notin> ?P'"
        from choice[OF this] obtain f::"nat \<Rightarrow> nat" where mono: "\<forall>i. f i \<ge> i"
          and Q_seq: "\<forall>i. (s (f i), t (f i)) \<in> ?Q"
          using mchain and `P = ?Q \<union> ?P'` by auto
        let ?S = "(op \<rhd> \<union> rstep R)"
        from cond and `P = ?Q \<union> ?P'` and mchain
          have s_no_vars: "\<forall>i. \<not> is_var (s i)"
          and  t_no_vars: "\<forall>i. \<not> is_var (t i)" 
          and  t_undef: "\<forall>i. \<not> defined R (the (root (t i)))" by auto
        have ts_undef: "\<forall>i. \<not> defined R (the (root (t i \<cdot> \<sigma> i)))"
        proof
          fix i
          from t_no_vars[THEN spec, of i] and t_undef[THEN spec, of i]
            show "\<not> defined R (the (root (t i \<cdot> \<sigma> i)))" by auto
        qed
        let ?s = "\<lambda>i. ?\<pi> (s i \<cdot> \<sigma> i)"
        let ?t = "\<lambda>i. ?\<pi> (t i \<cdot> \<sigma> i)"
        have seq: "\<forall>i. (?s i, ?s (Suc i)) \<in> ?S^*"
        proof
          fix i
          from mchain have "(s i, t i) \<in> ?Q \<union> ?P'"
            unfolding `P = ?Q \<union> ?P'`[symmetric] by simp
          hence "(?s i, ?t i) \<in> ?S^*"
          proof
            assume "(s i, t i) \<in> ?Q"
            with strict have "(?\<pi> (s i), ?\<pi> (t i)) \<in> ?S^+" by auto
            from supt_rstep_trancl_stable[OF this, of "\<sigma> i"]
              have "(?s i, ?t i) \<in> ?S^+"
              using s_no_vars[THEN spec, of i] and t_no_vars[THEN spec, of i] by simp
            thus ?thesis by simp
          next
            assume "(s i, t i) \<in> ?P'"
            with weak have "?\<pi> (s i) = ?\<pi> (t i)" by auto
            hence "?\<pi> (s i) \<cdot> \<sigma> i = ?\<pi> (t i) \<cdot> \<sigma> i" by simp
            hence "?s i = ?t i"
              using s_no_vars[THEN spec, of i] and t_no_vars[THEN spec, of i] by simp
            thus ?thesis by simp
          qed
          moreover have "(?t i, ?s (Suc i)) \<in> ?S^*"
          proof -
            from mchain have sn: "SN_elt (rstep R) (t i \<cdot> \<sigma> i)" by (simp add: minimal_cond_def)
            from mchain have "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep R)^*" by simp
            from SN_elt_rsteps_proj_term[OF this sn ts_undef[THEN spec, of i]]
              show ?thesis using rtrancl_Un_subset[of "op \<rhd>" "rstep R"] by auto
          qed
          ultimately show "(?s i, ?s (Suc i)) \<in> ?S^*" by simp
        qed
        from iseq_imp_steps[OF seq] and mono
          have between: "\<forall>i. (?s i, ?s (f i)) \<in> ?S^*" by simp
      
        let ?s' = "\<lambda>i. ?\<pi> (s (shift_by f (Suc i)) \<cdot> \<sigma> (shift_by f (Suc i)))"
        have "\<forall>i. (?s i, ?s (Suc (f i))) \<in> ?S^+"
        proof
          fix i
          from Q_seq and strict have "(?\<pi> (s (f i)), ?\<pi> (t (f i))) \<in> ?S^+" by auto
          from supt_rstep_trancl_stable[OF this]
            have "(?\<pi> (s (f i) \<cdot> \<sigma> (f i)), ?\<pi> (t (f i) \<cdot> \<sigma> (f i))) \<in> ?S^+"
            using s_no_vars[THEN spec, of "f i"] t_no_vars[THEN spec, of "f i"] by simp
          with between[THEN spec, of i] have "(?s i, ?\<pi> (t (f i) \<cdot> \<sigma> (f i))) \<in> ?S^+" by simp
          moreover have "(?\<pi> (t (f i) \<cdot> \<sigma> (f i)), ?\<pi> (s (Suc (f i)) \<cdot> \<sigma> (Suc (f i)))) \<in> ?S^*"
          proof -
            from mchain
              have sn: "SN_elt (rstep R) (t (f i) \<cdot> \<sigma> (f i))" by (simp add: minimal_cond_def)
            from mchain
              have "(t (f i) \<cdot> \<sigma> (f i), s (Suc (f i)) \<cdot> \<sigma> (Suc (f i))) \<in> (rstep R)^*" by simp
            from SN_elt_rsteps_proj_term[OF this sn ts_undef[THEN spec, of "f i"], of p]
              show ?thesis
              using rtrancl_Un_subset[of "op \<rhd>" "rstep R"] by auto
          qed
          ultimately show "(?s i, ?s (Suc (f i))) \<in> ?S^+" by simp
        qed
        hence "\<forall>i. (?s' i, ?s' (Suc i)) \<in> ?S^+" by simp
        with refl[of "?s' 0"]
          have "\<exists>S. S 0 = ?s' 0 \<and> (\<forall>i. (S i, S (Suc i)) \<in> ?S^+)" by best
        hence "\<not> SN_elt (?S^+) (?s' 0)" unfolding SN_defs by simp
        moreover have "SN_elt (?S^+) (?s' 0)"
        proof -
          have "(?t 0, ?s' 0) \<in> (?S^+)^*"
          proof -
            from mchain have sn: "SN_elt (rstep R) (t 0 \<cdot> \<sigma> 0)" by (simp add: minimal_cond_def)
            from mchain have "(t 0 \<cdot> \<sigma> 0, s (Suc 0) \<cdot> \<sigma> (Suc 0)) \<in> (rstep R)^*" by simp
            from SN_elt_rsteps_proj_term[OF this sn ts_undef[THEN spec, of "0"], of p]
              have "(?t 0, ?s (Suc 0)) \<in> ?S^*"
              using rtrancl_Un_subset[of "op \<rhd>" "rstep R"] by auto
            moreover have "(?s (Suc 0), ?s' 0) \<in> ?S^*"
            proof -
              have "shift_by f (Suc 0) \<ge> Suc 0" by simp
              from iseq_imp_steps[OF seq this] show ?thesis by simp
            qed
            ultimately show ?thesis by simp
          qed
          moreover have "SN_elt (?S^+) (?t 0)"
          proof -
            from mchain have "SN_elt (rstep R) (t 0 \<cdot> \<sigma> 0)" by (simp add: minimal_cond_def)
            moreover have "(t 0 \<cdot> \<sigma> 0, ?t 0) \<in> op \<unrhd>" by (simp add: supteq_proj_term[to_set])
            ultimately have "SN_elt (rstep R) (?t 0)"
              using subterm_preserves_SN[of R "t 0 \<cdot> \<sigma> 0" "?t 0"] by auto
            from SN_elt_imp_SN_elt_trancl[OF SN_elt_rstep_imp_SN_elt_supt_union_rstep[OF this]]
              show ?thesis .
          qed
          ultimately show ?thesis by (rule steps_preserve_SN_elt)
        qed
        ultimately show ?thesis ..
      qed
    qed
  qed
qed

end
